﻿namespace Mud.Helpers
{
	using Clio.Utilities;
	using ff14bot;
	using ff14bot.Managers;
	using ff14bot.Navigation;
	using ff14bot.Objects;
	using Properties;
	using System;
	using System.Collections.Generic;
	using System.Linq;

	public class Waypoint
	{
		public bool IsHardPoint;
		public Vector3 Location;
		public string Name;
		private static DateTime lastHardpoint = DateTime.Now;

		public Waypoint(GameObject obj)
		{
			bool hard = IsValid(obj.Location) && DateTime.Now.Subtract(LastHardpoint) > TimeSpan.FromSeconds(5);
			if (hard) LastHardpoint = DateTime.Now;
			Name = obj.Name;
			Location = obj.Location;
			IsHardPoint = hard;
		}

		public Waypoint(GameObject obj, bool hardPoint)
		{
			Name = obj.Name;
			Location = obj.Location;
			IsHardPoint = hardPoint;
		}

		public static DateTime LastHardpoint
		{
			get => lastHardpoint; set => lastHardpoint = value;
		}

		public float Distance(Waypoint other)
		{
			return (Location.Distance3D(other.Location));
		}

		public Waypoint Ground()
		{
			Vector3 rayStart = new Vector3(Location.X, Location.Y + 0.25f, Location.Z);
			Vector3 rayEnd = new Vector3(Location.X, Location.Y - 10f, Location.Z);
			Vector3 hit;
			Vector3 distances;
			WorldManager.Raycast(rayStart, rayEnd, out hit, out distances);
			if (IsValid(hit)) Location = hit;
			return this;
		}

		public bool InLineOfSight(Waypoint other)
		{
			return InLineOfSight(other, 1) && InLineOfSight(other, 2) && InLineOfSight(other, 3);
		}

		public bool IsCloserThan(Waypoint other)
		{
			return (Core.Player.Location.Distance3D(Location) < Core.Player.Location.Distance3D(other.Location));
		}

		public override string ToString()
		{
			return Name + " - " + Location + " Hardpoint: " + IsHardPoint;
		}

		private static bool IsValid(Vector3 loc)
		{
			return !(loc.X == 0 && loc.Y == 0 && loc.Z == 0);
		}

		private bool InLineOfSight(Waypoint other, int offsetY)
		{
			Vector3 start = Core.Player.Location;
			start.Y = start.Y + offsetY;
			Vector3 end = other.Location;
			end.Y = end.Y + offsetY;
			Vector3 hit;
			Vector3 distances;
			WorldManager.Raycast(start, end, out hit, out distances);
			return !IsValid(hit);
		}
	}

	public class WaypointManager
	{
		public const int WAYPOINT_DISTANCE = 2;
		private static GameObject _MoveTarget;
		private static Queue<Waypoint> CurrentWaypoints = new Queue<Waypoint>();
		private static bool isNavigating;
		public static bool IsNavigating
		{
			get => isNavigating; set => isNavigating = value;
		}

		public static Waypoint Next
		{
			get
			{
				if (CurrentWaypoints.Count == 0) return null;
				Waypoint peek = CurrentWaypoints.Peek();
				// Don't Skip Hardpoints
				if (CurrentWaypoints.Count > 1)
				{
					Waypoint peek2 = CurrentWaypoints.ToArray()[1];
					if (peek2.IsCloserThan(peek) && peek2.InLineOfSight(new Waypoint(Core.Player)))
					{
						Logging.Write(LogColors.INFO, $"{Logging.NAVIGATOR} Skipping Waypoint: {0}", CurrentWaypoints.Dequeue());
						return Next;
					}
				}
				else if (CurrentWaypoints.Count > 0)
				{
					Logging.Write(LogColors.PRIMARY, "Next Waypoint: {0}", CurrentWaypoints.Peek());
				}
				if (peek.Distance(new Waypoint(Core.Player)) > 0.25) return peek;
				Logging.Write(LogColors.WARNING, $"{Logging.NAVIGATOR} Reached Waypoint: {0}", CurrentWaypoints.Dequeue());
				return Next;
			}
		}

		public static void Track()
		{
			GameObject moveTarget = GetMoveTarget();
			if (moveTarget != null)
			{
				if (CurrentWaypoints.Count > 0)
				{
					Waypoint lastWaypoint = CurrentWaypoints.Last();
					Waypoint nextWaypoint = new Waypoint(moveTarget);
					if (lastWaypoint.Distance(nextWaypoint) > WAYPOINT_DISTANCE) CurrentWaypoints.Enqueue(nextWaypoint.Ground());
				}
				else
				{
					Waypoint nextWaypoint = new Waypoint(moveTarget).Ground();
					CurrentWaypoints.Enqueue(nextWaypoint);
				}
			}
			Logging.Write(LogColors.INFO, "WaypointManager Tracking {0}: Waypoints - {1}, Next - {2}", moveTarget, CurrentWaypoints.Count, Next);
		}

		// Needs Work
		internal static bool IsStuck()
		{
			return IsNavigating && (!MovementManager.IsMoving || MovementManager.IsMoving && MovementManager.Speed < 2);
		}

		internal static void MoveToNext()
		{
			GameObject moveTarget = GetMoveTarget();
			if (GetMoveTarget() != null
				// Start Navigating If > Max range
				&& ((!IsNavigating
					&& Core.Player.Location.Distance3D(moveTarget.Location) > (float)Settings.Default.AUTO_MOVE_FOLLOW_RANGE_MAX)
					// Keep Navigating If > Min Range
					|| (IsNavigating
					&& (Core.Player.Location.Distance3D(moveTarget.Location) > (float)Settings.Default.AUTO_MOVE_FOLLOW_RANGE_MIN))
					|| (Core.Player.Location.Distance3D(moveTarget.Location) > ((float)Settings.Default.AUTO_MOVE_TARGET_RANGE + moveTarget.CombatReach)
					&& MudBase.IsValidEnemy(Core.Player.CurrentTarget)
					&& MudBase.InCombat)))
			{
				//IsNavigating = true;
				new Waypoint(Core.Player);
				Waypoint next = Next;
				// Use Gaia/Service Navigator If Enabled & > Some Yards From Next Waypoint OR Not In LOS Of Next Waypoint
				{
					//Logging.Write(LogColors.WARNING, $"{Logging.NAVIGATOR} MoveTowards: {0}", next);
					Navigator.PlayerMover.MoveTowards(next.Location);
				}
			}
			// Otherwise Clear Waypoints
			else
			{
				//IsNavigating = false;
				CurrentWaypoints.Clear();
			}
		}

		internal static void StopNavigating()
		{
			Logging.Write(LogColors.INFO, $"{Logging.NAVIGATOR} Stopping, no Waypoint");
			//IsNavigating = false;
			CurrentWaypoints.Clear();
			Navigator.PlayerMover.MoveStop();
		}

		private static GameObject GetMoveTarget()
		{
			GameObject newTarget = null;
			IEnumerable<Character> targetTanks = MudBase.GetVisiblePartyMembers().Where(p => MudBase.IsTank(p)
				&& p.Location.Distance3D(Core.Player.Location) >= (float)Settings.Default.AUTO_MOVE_FOLLOW_RANGE_MIN
				&& Core.Player.Location.Distance3D(p.Location) <= (float)Settings.Default.MAX_TARGET_DISTANCE);
			if (Core.Player.HasTarget
					&& Core.Player.CurrentTarget.Location.Distance3D(Core.Player.Location) >= ((float)Settings.Default.AUTO_MOVE_TARGET_RANGE + Core.Player.CurrentTarget.CombatReach)
					&& Core.Player.Location.Distance3D(Core.Player.CurrentTarget.Location) <= (float)Settings.Default.MAX_TARGET_DISTANCE
					&& ((MudBase.IsValidEnemy(Core.Player.CurrentTarget)
					&& Core.Player.CurrentTarget.Location.Distance3D(Core.Player.Location) >= ((float)Settings.Default.AUTO_MOVE_TARGET_RANGE + Core.Player.CurrentTarget.CombatReach)
					&& (MudBase.MovementModes[Settings.Default.SELECTED_MOVEMENT_MODE].Equals("Combat")
					|| (MudBase.MovementModes[Settings.Default.SELECTED_MOVEMENT_MODE].Equals("Tank") && MudBase.InCombat)))
					|| (!MudBase.IsValidEnemy(Core.Player.CurrentTarget)
					&& Core.Player.CurrentTarget.Location.Distance3D(Core.Player.Location) >= (float)Settings.Default.AUTO_MOVE_FOLLOW_RANGE_MIN
					&& MudBase.MovementModes[Settings.Default.SELECTED_MOVEMENT_MODE].Equals("Follow"))))
			{
				newTarget = Core.Player.CurrentTarget;
				Logging.Write(LogColors.WARNING, $"{Logging.NAVIGATOR} New Target: {0}", newTarget);
			}
			else if (targetTanks.Count() > 0 && MudBase.MovementModes[Settings.Default.SELECTED_MOVEMENT_MODE].Equals("Tank"))
			{
				newTarget = targetTanks.First();
			}

			if (_MoveTarget != newTarget)
			{
				GameObject oldTarget = _MoveTarget;
				_MoveTarget = newTarget;
				OnTargetChanged(oldTarget, newTarget);
			}
			return _MoveTarget;
		}

		private static void OnTargetChanged(GameObject oldTarget, GameObject newTarget)
		{
			CurrentWaypoints.Clear();
			Track();
		}
	}
}